@@ -1,38 +1,53 @@
-import os
-import shutil
-from tempfile import mkstemp
-from tests import TestCase, add
+# -*- coding: utf-8 -*-
 
-from mutagen._compat import PY3, text_type, PY2
+import os
+import warnings
+
+from mutagen._compat import PY3, text_type, PY2, izip, cBytesIO
 from mutagen.asf import ASF, ASFHeaderError, ASFValue, UNICODE, DWORD, QWORD
 from mutagen.asf import BOOL, WORD, BYTEARRAY, GUID
+from mutagen.asf._util import guid2bytes, bytes2guid
+from mutagen.asf._objects import ContentDescriptionObject, \
+    ExtendedContentDescriptionObject, HeaderExtensionObject, \
+    MetadataObject, MetadataLibraryObject, CodecListObject, PaddingObject, \
+    HeaderObject
 from mutagen.asf import ASFUnicodeAttribute, ASFError, ASFByteArrayAttribute, \
     ASFBoolAttribute, ASFDWordAttribute, ASFQWordAttribute, ASFWordAttribute, \
     ASFGUIDAttribute
 
+from tests import TestCase, DATA_DIR, get_temp_copy
+
 
 class TASFFile(TestCase):
 
     def test_not_my_file(self):
         self.failUnlessRaises(
             ASFHeaderError, ASF,
-            os.path.join("tests", "data", "empty.ogg"))
+            os.path.join(DATA_DIR, "empty.ogg"))
         self.failUnlessRaises(
             ASFHeaderError, ASF,
-            os.path.join("tests", "data", "click.mpc"))
+            os.path.join(DATA_DIR, "click.mpc"))
 
-add(TASFFile)
+
+class TASFMisc(TestCase):
+
+    def test_guid(self):
+        ex = "75B22633-668E-11CF-A6D9-00AA0062CE6C"
+        b = guid2bytes(ex)
+        self.assertEqual(len(b), 16)
+        self.assertTrue(isinstance(b, bytes))
+        self.assertEqual(bytes2guid(b), ex)
 
 
 class TASFInfo(TestCase):
 
     def setUp(self):
         # WMA 9.1 64kbps CBR 48khz
-        self.wma1 = ASF(os.path.join("tests", "data", "silence-1.wma"))
+        self.wma1 = ASF(os.path.join(DATA_DIR, "silence-1.wma"))
         # WMA 9.1 Professional 192kbps VBR 44khz
-        self.wma2 = ASF(os.path.join("tests", "data", "silence-2.wma"))
+        self.wma2 = ASF(os.path.join(DATA_DIR, "silence-2.wma"))
         # WMA 9.1 Lossless 44khz
-        self.wma3 = ASF(os.path.join("tests", "data", "silence-3.wma"))
+        self.wma3 = ASF(os.path.join(DATA_DIR, "silence-3.wma"))
 
     def test_length(self):
         self.failUnlessAlmostEqual(self.wma1.info.length, 3.7, 1)
@@ -54,20 +69,59 @@ class TASFInfo(TestCase):
         self.failUnlessEqual(self.wma2.info.channels, 2)
         self.failUnlessEqual(self.wma3.info.channels, 2)
 
-add(TASFInfo)
+    def test_codec_type(self):
+        self.assertEqual(self.wma1.info.codec_type,
+                         "Windows Media Audio 9 Standard")
+        self.assertEqual(self.wma2.info.codec_type,
+                         "Windows Media Audio 9 Professional")
+        self.assertEqual(self.wma3.info.codec_type,
+                         "Windows Media Audio 9 Lossless")
+
+    def test_codec_name(self):
+        self.assertEqual(self.wma1.info.codec_name,
+            "Windows Media Audio 9.1")
+        self.assertEqual(self.wma2.info.codec_name,
+            "Windows Media Audio 9.1 Professional")
+        self.assertEqual(self.wma3.info.codec_name,
+            "Windows Media Audio 9.1 Lossless")
+
+    def test_codec_description(self):
+        self.assertEqual(self.wma1.info.codec_description,
+            "64 kbps, 48 kHz, stereo 2-pass CBR")
+        self.assertEqual(self.wma2.info.codec_description,
+            "192 kbps, 44 kHz, 2 channel 24 bit 2-pass VBR")
+        self.assertEqual(self.wma3.info.codec_description,
+            "VBR Quality 100, 44 kHz, 2 channel 16 bit 1-pass VBR")
+
+    def test_pprint(self):
+        self.assertTrue(self.wma1.info.pprint())
+        self.assertTrue(isinstance(self.wma1.info.pprint(), text_type))
 
 
 class TASF(TestCase):
 
     def setUp(self):
-        fd, self.filename = mkstemp(suffix='wma')
-        os.close(fd)
-        shutil.copy(self.original, self.filename)
+        self.filename = get_temp_copy(self.original)
         self.audio = ASF(self.filename)
 
     def tearDown(self):
         os.unlink(self.filename)
 
+
+class TASFMixin(object):
+
+    def test_header_object_misc(self):
+        header = self.audio._header
+        header.pprint()
+        repr(header)
+
+    def test_delete(self):
+        self.audio["QL/Bla"] = u"Foooooooooooooooooo"
+        self.audio.save(padding=lambda x: 0)
+        filesize = os.path.getsize(self.audio.filename)
+        self.audio.delete()
+        self.assertTrue(os.path.getsize(self.audio.filename) < filesize)
+
     def test_pprint(self):
         self.failUnless(self.audio.pprint())
 
@@ -81,7 +135,7 @@ class TASF(TestCase):
         self.failUnless(key in self.audio.tags.as_dict().keys())
         newvalue = self.audio[key]
         if isinstance(newvalue, list):
-            for a, b in zip(sorted(newvalue), sorted(result or value)):
+            for a, b in izip(sorted(newvalue), sorted(result or value)):
                 self.failUnlessEqual(a, b)
         else:
             self.failUnlessEqual(self.audio[key], result or value)
@@ -183,11 +237,11 @@ class TASF(TestCase):
         self.audio["QL/OneHasStream"] = [
             ASFValue("Whee", UNICODE, stream=2),
             ASFValue("Whee", UNICODE),
-            ]
+        ]
         self.audio["QL/AllHaveStream"] = [
             ASFValue("Whee", UNICODE, stream=1),
             ASFValue("Whee", UNICODE, stream=2),
-            ]
+        ]
         self.audio["QL/NoStream"] = ASFValue("Whee", UNICODE)
         self.audio.save()
         self.audio = ASF(self.audio.filename)
@@ -203,11 +257,11 @@ class TASF(TestCase):
         self.audio["QL/OneHasLang"] = [
             ASFValue("Whee", UNICODE, language=2),
             ASFValue("Whee", UNICODE),
-            ]
+        ]
         self.audio["QL/AllHaveLang"] = [
             ASFValue("Whee", UNICODE, language=1),
             ASFValue("Whee", UNICODE, language=2),
-            ]
+        ]
         self.audio["QL/NoLang"] = ASFValue("Whee", UNICODE)
         self.audio.save()
         self.audio = ASF(self.audio.filename)
@@ -223,7 +277,7 @@ class TASF(TestCase):
             ASFValue("Whee", UNICODE, language=2),
             ASFValue("Whee", UNICODE, stream=3, language=4),
             ASFValue("Whee", UNICODE),
-            ]
+        ]
         self.audio.save()
         self.audio = ASF(self.audio.filename)
         # order not preserved here because they end up in different objects.
@@ -254,6 +308,8 @@ class TASFAttributes(TestCase):
         self.assertRaises(ASFError, ASFUnicodeAttribute, data=b"\x00")
         self.assertEqual(ASFUnicodeAttribute(u"foo").value, u"foo")
 
+        assert ASFUnicodeAttribute(data=b"") == u""
+
     def test_ASFUnicodeAttribute_dunder(self):
         attr = ASFUnicodeAttribute(u"foo")
 
@@ -358,31 +414,24 @@ class TASFAttributes(TestCase):
         self.assertEqual(repr(attr).replace("0L", "0"), "ASFQWordAttribute(0)")
         self.assertEqual(int(attr), 0)
 
-add(TASFAttributes)
+
+class TASFTags1(TASF, TASFMixin):
+    original = os.path.join(DATA_DIR, "silence-1.wma")
 
 
-class TASFTags1(TASF):
-    original = os.path.join("tests", "data", "silence-1.wma")
-add(TASFTags1)
+class TASFTags2(TASF, TASFMixin):
+    original = os.path.join(DATA_DIR, "silence-2.wma")
 
 
-class TASFTags2(TASF):
-    original = os.path.join("tests", "data", "silence-2.wma")
-add(TASFTags2)
-
-
-class TASFTags3(TASF):
-    original = os.path.join("tests", "data", "silence-3.wma")
-add(TASFTags3)
+class TASFTags3(TASF, TASFMixin):
+    original = os.path.join(DATA_DIR, "silence-3.wma")
 
 
 class TASFIssue29(TestCase):
-    original = os.path.join("tests", "data", "issue_29.wma")
+    original = os.path.join(DATA_DIR, "issue_29.wma")
 
     def setUp(self):
-        fd, self.filename = mkstemp(suffix='wma')
-        os.close(fd)
-        shutil.copy(self.original, self.filename)
+        self.filename = get_temp_copy(self.original)
         self.audio = ASF(self.filename)
 
     def tearDown(self):
@@ -402,17 +451,29 @@ class TASFIssue29(TestCase):
         audio.save()
         audio = ASF(self.filename)
         self.failIf("Description" in audio)
-add(TASFIssue29)
+
+
+class TASFObjects(TestCase):
+
+    filename = os.path.join(DATA_DIR, "silence-1.wma")
+
+    def test_invalid_header(self):
+        with warnings.catch_warnings():
+            warnings.simplefilter("ignore")
+            asf = ASF()
+        fileobj = cBytesIO(
+            b"0&\xb2u\x8ef\xcf\x11\xa6\xd9\x00\xaa\x00b\xcel\x19\xbf\x01\x00"
+            b"\x00\x00\x00\x00\x07\x00\x00\x00\x01\x02")
+        self.assertRaises(
+            ASFHeaderError, HeaderObject.parse_full, asf, fileobj)
 
 
 class TASFAttrDest(TestCase):
 
-    original = os.path.join("tests", "data", "silence-1.wma")
+    original = os.path.join(DATA_DIR, "silence-1.wma")
 
     def setUp(self):
-        fd, self.filename = mkstemp(suffix='wma')
-        os.close(fd)
-        shutil.copy(self.original, self.filename)
+        self.filename = get_temp_copy(self.original)
         audio = ASF(self.filename)
         audio.clear()
         audio.save()
@@ -482,17 +543,13 @@ class TASFAttrDest(TestCase):
         self.assertEqual(new["Title"], [u""])
         self.assertFalse("Copyright" in new)
 
-add(TASFAttrDest)
-
 
 class TASFLargeValue(TestCase):
 
-    original = os.path.join("tests", "data", "silence-1.wma")
+    original = os.path.join(DATA_DIR, "silence-1.wma")
 
     def setUp(self):
-        fd, self.filename = mkstemp(suffix='wma')
-        os.close(fd)
-        shutil.copy(self.original, self.filename)
+        self.filename = get_temp_copy(self.original)
 
     def tearDown(self):
         os.unlink(self.filename)
@@ -532,7 +589,7 @@ class TASFLargeValue(TestCase):
         self.failIf("QL/LargeObject" not in dict(audio.to_metadata_library))
 
     def test_save_guid(self):
-        # http://code.google.com/p/mutagen/issues/detail?id=81
+        # https://github.com/quodlibet/mutagen/issues/81
         audio = ASF(self.filename)
         audio["QL/GuidObject"] = [ASFValue(b" " * 16, GUID)]
         audio.save()
@@ -540,29 +597,63 @@ class TASFLargeValue(TestCase):
         self.failIf("QL/GuidObject" in audio.to_metadata)
         self.failIf("QL/GuidObject" not in dict(audio.to_metadata_library))
 
-add(TASFLargeValue)
 
+class TASFSave(TestCase):
+    # https://github.com/quodlibet/mutagen/issues/81#issuecomment-207014936
 
-class TASFUpdateSize(TestCase):
-    # http://code.google.com/p/mutagen/issues/detail?id=81#c4
-
-    original = os.path.join("tests", "data", "silence-1.wma")
+    original = os.path.join(DATA_DIR, "silence-1.wma")
 
     def setUp(self):
-        fd, self.filename = mkstemp(suffix='wma')
-        os.close(fd)
-        shutil.copy(self.original, self.filename)
-        audio = ASF(self.filename)
-        audio["large_value1"] = "#" * 50000
-        audio.save()
+        self.filename = get_temp_copy(self.original)
+        self.audio = ASF(self.filename)
 
     def tearDown(self):
         os.unlink(self.filename)
 
+    def test_save_filename(self):
+        self.audio.save(self.audio.filename)
+
     def test_multiple_delete(self):
+        self.audio["large_value1"] = "#" * 50000
+        self.audio.save()
+
         audio = ASF(self.filename)
         for tag in audio.keys():
             del(audio[tag])
             audio.save()
 
-add(TASFUpdateSize)
+    def test_readd_objects(self):
+        header = self.audio._header
+        del header.objects[:]
+        self.audio.save()
+        self.assertTrue(header.get_child(ContentDescriptionObject.GUID))
+        self.assertTrue(
+            header.get_child(ExtendedContentDescriptionObject.GUID))
+        self.assertTrue(header.get_child(HeaderExtensionObject.GUID))
+        ext = header.get_child(HeaderExtensionObject.GUID)
+        self.assertTrue(ext.get_child(MetadataObject.GUID))
+        self.assertTrue(ext.get_child(MetadataLibraryObject.GUID))
+
+    def test_keep_others(self):
+        self.audio.save()
+        new = ASF(self.filename)
+        self.assertTrue(new._header.get_child(CodecListObject.GUID))
+
+    def test_padding(self):
+        old_tags = sorted(self.audio.items())
+
+        def get_padding(fn):
+            header = ASF(fn)._header
+            return len(header.get_child(PaddingObject.GUID).data)
+
+        for i in [0, 1, 2, 3, 42, 100, 5000, 30432, 1]:
+
+            def padding_cb(info):
+                self.assertEqual(info.size, 30432)
+                return i
+
+            self.audio.save(padding=padding_cb)
+            self.assertEqual(get_padding(self.filename), i)
+
+        last = ASF(self.filename)
+        self.assertEqual(sorted(last.items()), old_tags)
