@@ -1,15 +1,17 @@
-import os
-import shutil
-import struct
+# -*- coding: utf-8 -*-
 
-from mutagen._compat import cBytesIO, PY3
-from tempfile import mkstemp
-from tests import TestCase, add
+import os
+import struct
+import subprocess
+
+from mutagen._compat import cBytesIO, PY3, text_type, PY2, izip
+from tests import TestCase, DATA_DIR, get_temp_copy
 from mutagen.mp4 import (MP4, Atom, Atoms, MP4Tags, MP4Info, delete, MP4Cover,
                          MP4MetadataError, MP4FreeForm, error, AtomDataType,
-                         MP4MetadataValueError, AudioSampleEntry)
+                         AtomError, _item_sort_key, MP4StreamInfoError)
+from mutagen.mp4._util import parse_full_atom
+from mutagen.mp4._as_entry import AudioSampleEntry, ASEntryError
 from mutagen._util import cdata
-from os import devnull
 
 
 class TAtom(TestCase):
@@ -22,16 +24,22 @@ class TAtom(TestCase):
     def test_length_1(self):
         fileobj = cBytesIO(b"\x00\x00\x00\x01atom"
                            b"\x00\x00\x00\x00\x00\x00\x00\x10" + b"\x00" * 16)
-        self.failUnlessEqual(Atom(fileobj).length, 16)
+        atom = Atom(fileobj)
+        self.failUnlessEqual(atom.length, 16)
+        self.failUnlessEqual(atom.datalength, 0)
 
     def test_length_64bit_less_than_16(self):
         fileobj = cBytesIO(b"\x00\x00\x00\x01atom"
                            b"\x00\x00\x00\x00\x00\x00\x00\x08" + b"\x00" * 8)
-        self.assertRaises(error, Atom, fileobj)
+        self.assertRaises(AtomError, Atom, fileobj)
 
     def test_length_less_than_8(self):
         fileobj = cBytesIO(b"\x00\x00\x00\x02atom")
-        self.assertRaises(MP4MetadataError, Atom, fileobj)
+        self.assertRaises(AtomError, Atom, fileobj)
+
+    def test_truncated(self):
+        self.assertRaises(AtomError, Atom, cBytesIO(b"\x00"))
+        self.assertRaises(AtomError, Atom, cBytesIO(b"\x00\x00\x00\x01atom"))
 
     def test_render_too_big(self):
         class TooBig(bytes):
@@ -49,13 +57,14 @@ class TAtom(TestCase):
 
     def test_non_top_level_length_0_is_invalid(self):
         data = cBytesIO(struct.pack(">I4s", 0, b"whee"))
-        self.assertRaises(MP4MetadataError, Atom, data, level=1)
+        self.assertRaises(AtomError, Atom, data, level=1)
 
     def test_length_0(self):
         fileobj = cBytesIO(b"\x00\x00\x00\x00atom" + 40 * b"\x00")
         atom = Atom(fileobj)
         self.failUnlessEqual(fileobj.tell(), 48)
         self.failUnlessEqual(atom.length, 48)
+        self.failUnlessEqual(atom.datalength, 40)
 
     def test_length_0_container(self):
         data = cBytesIO(struct.pack(">I4s", 0, b"moov") +
@@ -80,14 +89,13 @@ class TAtom(TestCase):
         self.assertFalse(ok)
         self.assertEqual(data, payload)
 
-add(TAtom)
-
 
 class TAtoms(TestCase):
-    filename = os.path.join("tests", "data", "has-tags.m4a")
+    filename = os.path.join(DATA_DIR, "has-tags.m4a")
 
     def setUp(self):
-        self.atoms = Atoms(open(self.filename, "rb"))
+        with open(self.filename, "rb") as h:
+            self.atoms = Atoms(h)
 
     def test_getitem(self):
         self.failUnless(self.atoms[b"moov"])
@@ -114,14 +122,13 @@ class TAtoms(TestCase):
 
     def test_repr(self):
         repr(self.atoms)
-add(TAtoms)
 
 
 class TMP4Info(TestCase):
 
     def test_no_soun(self):
         self.failUnlessRaises(
-            IOError, self.test_mdhd_version_1, b"vide")
+            error, self.test_mdhd_version_1, b"vide")
 
     def test_mdhd_version_1(self, soun=b"soun"):
         mdhd = Atom.render(b"mdhd", (b"\x01\x00\x00\x00" + b"\x00" * 16 +
@@ -151,7 +158,13 @@ class TMP4Info(TestCase):
         atoms = Atoms(fileobj)
         info = MP4Info(atoms, fileobj)
         self.failUnlessEqual(info.length, 8)
-add(TMP4Info)
+
+    def test_no_tracks(self):
+        moov = Atom.render(b"moov", b"")
+        fileobj = cBytesIO(moov)
+        atoms = Atoms(fileobj)
+        with self.assertRaises(MP4StreamInfoError):
+            MP4Info(atoms, fileobj)
 
 
 class TMP4Tags(TestCase):
@@ -166,7 +179,7 @@ class TMP4Tags(TestCase):
     def test_parse_multiple_atoms(self):
         # while we don't write multiple values as multiple atoms
         # still read them
-        # https://bitbucket.org/lazka/mutagen/issue/165
+        # https://github.com/quodlibet/mutagen/issues/165
         data = Atom.render(b"data", b"\x00\x00\x00\x01" + b"\x00" * 4 + b"foo")
         grp1 = Atom.render(b"\xa9grp", data)
         data = Atom.render(b"data", b"\x00\x00\x00\x01" + b"\x00" * 4 + b"bar")
@@ -259,6 +272,17 @@ class TMP4Tags(TestCase):
             b"\x00\x00\x00\x15\x00\x00\x00\x00\x00"
         )
 
+    def test_render_integer_min_size(self):
+        render_int = MP4Tags()._MP4Tags__render_integer
+
+        data = render_int('stik', [42], 1)
+        tags = self.wrap_ilst(data)
+        assert tags['stik'] == [42]
+
+        assert len(render_int('stik', [42], 2)) == len(data) + 1
+        assert len(render_int('stik', [42], 4)) == len(data) + 3
+        assert len(render_int('stik', [42], 8)) == len(data) + 7
+
     def test_render_text(self):
         self.failUnlessEqual(
             MP4Tags()._MP4Tags__render_text(
@@ -304,6 +328,16 @@ class TMP4Tags(TestCase):
         tags = self.wrap_ilst(data)
         self.assertFalse(tags)
 
+    def test_parse_tmpo(self):
+        for d, v in [(b"\x01", 1), (b"\x01\x02", 258),
+                     (b"\x01\x02\x03", 66051), (b"\x01\x02\x03\x04", 16909060),
+                     (b"\x01\x02\x03\x04\x05\x06\x07\x08", 72623859790382856)]:
+            data = Atom.render(
+                b"data", b"\x00\x00\x00\x15" + b"\x00\x00\x00\x00" + d)
+            tmpo = Atom.render(b"tmpo", data)
+            tags = self.wrap_ilst(tmpo)
+            assert tags["tmpo"][0] == v
+
     def test_write_back_bad_atoms(self):
         # write a broken atom and try to load it
         data = Atom.render(b"datA", b"\x00\x00\x00\x01\x00\x00\x00\x00wheeee")
@@ -312,31 +346,33 @@ class TMP4Tags(TestCase):
         self.assertFalse(tags)
 
         # save it into an existing mp4
-        original = os.path.join("tests", "data", "has-tags.m4a")
-        fd, filename = mkstemp(suffix='.mp4')
-        os.close(fd)
-        shutil.copy(original, filename)
-        delete(filename)
+        original = os.path.join(DATA_DIR, "has-tags.m4a")
+        filename = get_temp_copy(original)
+        try:
+            delete(filename)
 
-        # it should still end up in the file
-        tags.save(filename)
-        with open(filename, "rb") as h:
-            self.assertTrue(b"wheeee" in h.read())
+            # it should still end up in the file
+            tags.save(filename)
+            with open(filename, "rb") as h:
+                self.assertTrue(b"wheeee" in h.read())
 
-        # if we define our own aART throw away the broken one
-        tags["aART"] = ["new"]
-        tags.save(filename)
-        with open(filename, "rb") as h:
-            self.assertFalse(b"wheeee" in h.read())
+            # if we define our own aART throw away the broken one
+            tags["aART"] = ["new"]
+            tags.save(filename)
+            with open(filename, "rb") as h:
+                self.assertFalse(b"wheeee" in h.read())
 
-        # add the broken one back and delete all tags including the broken one
-        del tags["aART"]
-        tags.save(filename)
-        with open(filename, "rb") as h:
-            self.assertTrue(b"wheeee" in h.read())
-        delete(filename)
-        with open(filename, "rb") as h:
-            self.assertFalse(b"wheeee" in h.read())
+            # add the broken one back and delete all tags including
+            # the broken one
+            del tags["aART"]
+            tags.save(filename)
+            with open(filename, "rb") as h:
+                self.assertTrue(b"wheeee" in h.read())
+            delete(filename)
+            with open(filename, "rb") as h:
+                self.assertFalse(b"wheeee" in h.read())
+        finally:
+            os.unlink(filename)
 
     def test_render_freeform(self):
         data = (
@@ -396,11 +432,11 @@ class TMP4Tags(TestCase):
     def test_pprint_non_text_list(self):
         tags = MP4Tags()
         tags["tmpo"] = [120, 121]
-        tags["trck"] = [(1, 2), (3, 4)]
+        tags["trkn"] = [(1, 2), (3, 4)]
         tags.pprint()
 
     def test_freeform_data(self):
-        # http://code.google.com/p/mutagen/issues/detail?id=103
+        # https://github.com/quodlibet/mutagen/issues/103
         key = "----:com.apple.iTunes:Encoding Params"
         value = (b"vers\x00\x00\x00\x01acbf\x00\x00\x00\x01brat\x00\x01\xf4"
                  b"\x00cdcv\x00\x01\x05\x04")
@@ -423,22 +459,26 @@ class TMP4Tags(TestCase):
         v = self.wrap_ilst(data)[key][0]
         self.failUnlessEqual(v.dataformat, AtomDataType.UTF8)
 
-add(TMP4Tags)
-
 
 class TMP4(TestCase):
+
     def setUp(self):
-        fd, self.filename = mkstemp(suffix='.m4a')
-        os.close(fd)
-        shutil.copy(self.original, self.filename)
+        self.filename = get_temp_copy(self.original)
         self.audio = MP4(self.filename)
 
+    def tearDown(self):
+        os.unlink(self.filename)
+
+
+class TMP4Mixin(object):
+
     def faad(self):
         if not have_faad:
             return
-        value = os.system("faad %s -o %s > %s 2> %s" % (self.filename, devnull,
-                                                        devnull, devnull))
-        self.failIf(value and value != NOTFOUND)
+        self.assertEqual(call_faad("-w", self.filename), 0)
+
+    def test_set_inval(self):
+        self.assertRaises(TypeError, self.audio.__setitem__, "\xa9nam", 42)
 
     def test_score(self):
         fileobj = open(self.filename, "rb")
@@ -461,6 +501,9 @@ class TMP4(TestCase):
     def test_length(self):
         self.failUnlessAlmostEqual(3.7, self.audio.info.length, 1)
 
+    def test_kind(self):
+        self.assertEqual(self.audio.info.codec, u'mp4a.40.2')
+
     def test_padding(self):
         self.audio["\xa9nam"] = u"wheeee" * 10
         self.audio.save()
@@ -473,35 +516,37 @@ class TMP4(TestCase):
     def test_padding_2(self):
         self.audio["\xa9nam"] = u"wheeee" * 10
         self.audio.save()
+
         # Reorder "free" and "ilst" atoms
-        fileobj = open(self.audio.filename, "rb+")
-        atoms = Atoms(fileobj)
-        meta = atoms[b"moov", b"udta", b"meta"]
-        meta_length1 = meta.length
-        ilst = meta[b"ilst", ]
-        free = meta[b"free", ]
-        self.failUnlessEqual(ilst.offset + ilst.length, free.offset)
-        fileobj.seek(ilst.offset)
-        ilst_data = fileobj.read(ilst.length)
-        fileobj.seek(free.offset)
-        free_data = fileobj.read(free.length)
-        fileobj.seek(ilst.offset)
-        fileobj.write(free_data + ilst_data)
-        fileobj.close()
-        fileobj = open(self.audio.filename, "rb+")
-        atoms = Atoms(fileobj)
-        meta = atoms[b"moov", b"udta", b"meta"]
-        ilst = meta[b"ilst", ]
-        free = meta[b"free", ]
-        self.failUnlessEqual(free.offset + free.length, ilst.offset)
-        fileobj.close()
+        with open(self.audio.filename, "rb+") as fileobj:
+            atoms = Atoms(fileobj)
+            meta = atoms[b"moov", b"udta", b"meta"]
+            meta_length1 = meta.length
+            ilst = meta[b"ilst", ]
+            free = meta[b"free", ]
+            self.failUnlessEqual(ilst.offset + ilst.length, free.offset)
+            fileobj.seek(ilst.offset)
+            ilst_data = fileobj.read(ilst.length)
+            fileobj.seek(free.offset)
+            free_data = fileobj.read(free.length)
+            fileobj.seek(ilst.offset)
+            fileobj.write(free_data + ilst_data)
+
+        with open(self.audio.filename, "rb+") as fileobj:
+            atoms = Atoms(fileobj)
+            meta = atoms[b"moov", b"udta", b"meta"]
+            ilst = meta[b"ilst", ]
+            free = meta[b"free", ]
+            self.failUnlessEqual(free.offset + free.length, ilst.offset)
+
         # Save the file
         self.audio["\xa9nam"] = u"wheeee" * 11
         self.audio.save()
+
         # Check the order of "free" and "ilst" atoms
-        fileobj = open(self.audio.filename, "rb+")
-        atoms = Atoms(fileobj)
-        fileobj.close()
+        with open(self.audio.filename, "rb+") as fileobj:
+            atoms = Atoms(fileobj)
+
         meta = atoms[b"moov", b"udta", b"meta"]
         ilst = meta[b"ilst", ]
         free = meta[b"free", ]
@@ -521,7 +566,7 @@ class TMP4(TestCase):
         try:
             self.set_key('\xa9nam', [b'\xe3\x82\x8a\xe3\x81\x8b'],
                          result=[u'\u308a\u304b'])
-        except MP4MetadataValueError:
+        except TypeError:
             if not PY3:
                 raise
 
@@ -529,27 +574,9 @@ class TMP4(TestCase):
         self.set_key('----:net.sacredchao.Mutagen:test key',
                      [MP4FreeForm(b'woooo', 142, 42)])
 
-    def test_freeform_cmp(self):
-        self.assertReallyEqual(
-            MP4FreeForm(b'woooo', 142, 42), MP4FreeForm(b'woooo', 142, 42))
-        self.assertReallyNotEqual(
-            MP4FreeForm(b'woooo', 142, 43), MP4FreeForm(b'woooo', 142, 42))
-        self.assertReallyNotEqual(
-            MP4FreeForm(b'woooo', 143, 42), MP4FreeForm(b'woooo', 142, 42))
-        self.assertReallyNotEqual(
-            MP4FreeForm(b'wooox', 142, 42), MP4FreeForm(b'woooo', 142, 42))
-
-    def test_cover_cmp(self):
-        self.assertReallyEqual(
-            MP4Cover(b'woooo', 142), MP4Cover(b'woooo', 142))
-        self.assertReallyNotEqual(
-            MP4Cover(b'woooo', 143), MP4Cover(b'woooo', 142))
-        self.assertReallyNotEqual(
-            MP4Cover(b'woooo', 142), MP4Cover(b'wooox', 142))
-
     def test_invalid_text(self):
         self.assertRaises(
-            MP4MetadataValueError, self.set_key, '\xa9nam', [b'\xff'])
+            TypeError, self.audio.__setitem__, '\xa9nam', [b'\xff'])
 
     def test_save_text(self):
         self.set_key('\xa9nam', [u"Some test name"])
@@ -608,12 +635,39 @@ class TMP4(TestCase):
     def test_tempo(self):
         self.set_key('tmpo', [150])
         self.set_key('tmpo', [])
+        self.set_key('tmpo', [0])
+        self.set_key('tmpo', [cdata.int16_min])
+        self.set_key('tmpo', [cdata.int32_min])
+        self.set_key('tmpo', [cdata.int64_min])
+        self.set_key('tmpo', [cdata.int16_max])
+        self.set_key('tmpo', [cdata.int32_max])
+        self.set_key('tmpo', [cdata.int64_max])
+
+    def test_various_int(self):
+        keys = [
+            "stik", "rtng", "plID", "cnID", "geID", "atID", "sfID",
+            "cmID", "akID", "tvsn", "tves",
+        ]
+
+        for key in keys:
+            self.set_key(key, [])
+            self.set_key(key, [0])
+            self.set_key(key, [1])
+            self.set_key(key, [cdata.int64_max])
+
+    def test_movements(self):
+        self.set_key('shwm', [1])
+        self.set_key('\xa9mvc', [42])
+        self.set_key('\xa9mvi', [24])
+        self.set_key('\xa9mvn', [u"movement"])
+        self.set_key('\xa9wrk', [u"work"])
 
     def test_tempos(self):
         self.set_key('tmpo', [160, 200], faad=False)
 
     def test_tempo_invalid(self):
-        for badvalue in [[10000000], [-1], 10, "foo"]:
+        for badvalue in [
+                [cdata.int64_max + 1], [cdata.int64_min - 1], 10, "foo"]:
             self.failUnlessRaises(ValueError, self.set_key, 'tmpo', badvalue)
 
     def test_compilation(self):
@@ -645,16 +699,17 @@ class TMP4(TestCase):
 
     def test_podcast_url(self):
         self.set_key('purl', ['http://pdl.warnerbros.com/wbie/'
-                               'justiceleagueheroes/audio/JLH_EA.xml'])
+                              'justiceleagueheroes/audio/JLH_EA.xml'])
 
     def test_episode_guid(self):
         self.set_key('catg', ['falling-star-episode-1'])
 
     def test_pprint(self):
         self.failUnless(self.audio.pprint())
+        self.assertTrue(isinstance(self.audio.pprint(), text_type))
 
     def test_pprint_binary(self):
-        self.audio["covr"] = [b"\x00\xa9\garbage"]
+        self.audio["covr"] = [b"\x00\xa9garbage"]
         self.failUnless(self.audio.pprint())
 
     def test_pprint_pair(self):
@@ -715,17 +770,34 @@ class TMP4(TestCase):
         self.audio["\xa9nam"] = "wheeeeeeee"
         self.audio.save()
         bb = self.__read_offsets(self.filename)
-        for a, b in zip(aa, bb):
+        for a, b in izip(aa, bb):
             self.failUnlessEqual(a, b)
 
     def test_mime(self):
         self.failUnless("audio/mp4" in self.audio.mime)
 
-    def tearDown(self):
-        os.unlink(self.filename)
+    def test_set_init_padding_zero(self):
+        if self.audio.tags is None:
+            self.audio.add_tags()
+        self.audio.save(padding=lambda x: 0)
+        self.assertEqual(MP4(self.audio.filename)._padding, 0)
+
+    def test_set_init_padding_large(self):
+        if self.audio.tags is None:
+            self.audio.add_tags()
+        self.audio.save(padding=lambda x: 5000)
+        self.assertEqual(MP4(self.audio.filename)._padding, 5000)
+
+    def test_set_various_padding(self):
+        if self.audio.tags is None:
+            self.audio.add_tags()
+        for i in [0, 1, 2, 3, 1024, 983, 5000, 0, 1]:
+            self.audio.save(padding=lambda x: i)
+            self.assertEqual(MP4(self.audio.filename)._padding, i)
+            self.faad()
 
 
-class TMP4HasTags(TMP4):
+class TMP4HasTagsMixin(TMP4Mixin):
     def test_save_simple(self):
         self.audio.save()
         self.faad()
@@ -753,11 +825,19 @@ class TMP4HasTags(TMP4):
     def test_not_my_file(self):
         # should raise something like "Not a MP4 file"
         self.failUnlessRaisesRegexp(
-            error, "MP4", MP4, os.path.join("tests", "data", "empty.ogg"))
+            error, "MP4", MP4, os.path.join(DATA_DIR, "empty.ogg"))
+
+    def test_delete_remove_padding(self):
+        self.audio.clear()
+        self.audio.tags['foob'] = u"foo"
+        self.audio.save(padding=lambda x: 0)
+        filesize = os.path.getsize(self.audio.filename)
+        self.audio.delete()
+        self.assertTrue(os.path.getsize(self.audio.filename) < filesize)
 
 
-class TMP4Datatypes(TMP4HasTags):
-    original = os.path.join("tests", "data", "has-tags.m4a")
+class TMP4Datatypes(TMP4, TMP4HasTagsMixin):
+    original = os.path.join(DATA_DIR, "has-tags.m4a")
 
     def test_has_freeform(self):
         key = "----:com.apple.iTunes:iTunNORM"
@@ -773,12 +853,17 @@ class TMP4Datatypes(TMP4HasTags):
         self.failUnlessEqual(covr[0].imageformat, MP4Cover.FORMAT_PNG)
         self.failUnlessEqual(covr[1].imageformat, MP4Cover.FORMAT_JPEG)
 
-add(TMP4Datatypes)
+    def test_pprint(self):
+        text = self.audio.tags.pprint().splitlines()
+        self.assertTrue(u"©ART=Test Artist" in text)
+
+    def test_get_padding(self):
+        self.assertEqual(self.audio._padding, 1634)
 
 
-class TMP4CovrWithName(TMP4):
+class TMP4CovrWithName(TMP4, TMP4Mixin):
     # http://bugs.musicbrainz.org/ticket/5894
-    original = os.path.join("tests", "data", "covr-with-name.m4a")
+    original = os.path.join(DATA_DIR, "covr-with-name.m4a")
 
     def test_has_covr(self):
         self.failUnless('covr' in self.audio.tags)
@@ -787,11 +872,9 @@ class TMP4CovrWithName(TMP4):
         self.failUnlessEqual(covr[0].imageformat, MP4Cover.FORMAT_PNG)
         self.failUnlessEqual(covr[1].imageformat, MP4Cover.FORMAT_JPEG)
 
-add(TMP4CovrWithName)
 
-
-class TMP4HasTags64Bit(TMP4HasTags):
-    original = os.path.join("tests", "data", "truncated-64bit.mp4")
+class TMP4HasTags64Bit(TMP4, TMP4HasTagsMixin):
+    original = os.path.join(DATA_DIR, "truncated-64bit.mp4")
 
     def test_has_covr(self):
         pass
@@ -806,11 +889,9 @@ class TMP4HasTags64Bit(TMP4HasTags):
         # This is only half a file, so FAAD segfaults. Can't test. :(
         pass
 
-add(TMP4HasTags64Bit)
 
-
-class TMP4NoTagsM4A(TMP4):
-    original = os.path.join("tests", "data", "no-tags.m4a")
+class TMP4NoTagsM4A(TMP4, TMP4Mixin):
+    original = os.path.join(DATA_DIR, "no-tags.m4a")
 
     def test_no_tags(self):
         self.failUnless(self.audio.tags is None)
@@ -819,11 +900,9 @@ class TMP4NoTagsM4A(TMP4):
         self.audio.add_tags()
         self.failUnlessRaises(error, self.audio.add_tags)
 
-add(TMP4NoTagsM4A)
 
-
-class TMP4NoTags3G2(TMP4):
-    original = os.path.join("tests", "data", "no-tags.3g2")
+class TMP4NoTags3G2(TMP4, TMP4Mixin):
+    original = os.path.join(DATA_DIR, "no-tags.3g2")
 
     def test_no_tags(self):
         self.failUnless(self.audio.tags is None)
@@ -837,16 +916,12 @@ class TMP4NoTags3G2(TMP4):
     def test_length(self):
         self.failUnlessAlmostEqual(15, self.audio.info.length, 1)
 
-add(TMP4NoTags3G2)
-
 
 class TMP4UpdateParents64Bit(TestCase):
-    original = os.path.join("tests", "data", "64bit.mp4")
+    original = os.path.join(DATA_DIR, "64bit.mp4")
 
     def setUp(self):
-        fd, self.filename = mkstemp(suffix='.mp4')
-        os.close(fd)
-        shutil.copy(self.original, self.filename)
+        self.filename = get_temp_copy(self.original)
 
     def test_update_parents(self):
         with open(self.filename, "rb") as fileobj:
@@ -855,22 +930,20 @@ class TMP4UpdateParents64Bit(TestCase):
             self.assertEqual(61, atoms.atoms[0].children[0].length)
             tags = MP4Tags(atoms, fileobj)
             tags['pgap'] = True
-            tags.save(self.filename)
+            tags.save(self.filename, padding=lambda x: 0)
 
         with open(self.filename, "rb") as fileobj:
             atoms = Atoms(fileobj)
             # original size + 'pgap' size + padding
-            self.assertEqual(77 + 25 + 974, atoms.atoms[0].length)
-            self.assertEqual(61 + 25 + 974, atoms.atoms[0].children[0].length)
+            self.assertEqual(77 + 25 + 8, atoms.atoms[0].length)
+            self.assertEqual(61 + 25 + 8, atoms.atoms[0].children[0].length)
 
     def tearDown(self):
         os.unlink(self.filename)
 
-add(TMP4UpdateParents64Bit)
-
 
 class TMP4ALAC(TestCase):
-    original = os.path.join("tests", "data", "alac.m4a")
+    original = os.path.join(DATA_DIR, "alac.m4a")
 
     def setUp(self):
         self.audio = MP4(self.original)
@@ -890,7 +963,81 @@ class TMP4ALAC(TestCase):
     def test_bitrate(self):
         self.assertEqual(self.audio.info.bitrate, 2764)
 
-add(TMP4ALAC)
+    def test_kind(self):
+        self.assertEqual(self.audio.info.codec, u'alac')
+
+
+class TMP4Misc(TestCase):
+
+    def test_no_audio_tracks(self):
+        data = Atom.render(b"moov", Atom.render(b"udta", b""))
+        fileobj = cBytesIO(data)
+        audio = MP4(fileobj)
+        assert audio.info
+        assert audio.pprint()
+        info = audio.info
+        assert isinstance(info.bitrate, int)
+        assert isinstance(info.length, float)
+        assert isinstance(info.channels, int)
+        assert isinstance(info.sample_rate, int)
+        assert isinstance(info.bits_per_sample, int)
+        assert isinstance(info.codec, text_type)
+        assert isinstance(info.codec_description, text_type)
+
+    def test_parse_full_atom(self):
+        p = parse_full_atom(b"\x01\x02\x03\x04\xff")
+        self.assertEqual(p, (1, 131844, b'\xff'))
+
+        self.assertRaises(ValueError, parse_full_atom, b"\x00\x00\x00")
+
+    def test_sort_items(self):
+        items = [
+            ("\xa9nam", ["foo"]),
+            ("gnre", ["fo"]),
+            ("----", ["123"]),
+            ("----", ["1234"]),
+        ]
+
+        sorted_items = sorted(items, key=lambda kv: _item_sort_key(*kv))
+        self.assertEqual(sorted_items, items)
+
+
+class TMP4Freeform(TestCase):
+
+    def test_cmp(self):
+        self.assertReallyEqual(
+            MP4FreeForm(b'woooo', 142, 42), MP4FreeForm(b'woooo', 142, 42))
+        self.assertReallyNotEqual(
+            MP4FreeForm(b'woooo', 142, 43), MP4FreeForm(b'woooo', 142, 42))
+        self.assertReallyNotEqual(
+            MP4FreeForm(b'woooo', 143, 42), MP4FreeForm(b'woooo', 142, 42))
+        self.assertReallyNotEqual(
+            MP4FreeForm(b'wooox', 142, 42), MP4FreeForm(b'woooo', 142, 42))
+
+    def test_cmp_bytes(self):
+        self.assertReallyEqual(MP4FreeForm(b'woooo'), b"woooo")
+        self.assertReallyNotEqual(MP4FreeForm(b'woooo'), b"foo")
+        if PY2:
+            self.assertReallyEqual(MP4FreeForm(b'woooo'), u"woooo")
+            self.assertReallyNotEqual(MP4FreeForm(b'woooo'), u"foo")
+
+
+class TMP4Cover(TestCase):
+
+    def test_cmp(self):
+        self.assertReallyEqual(
+            MP4Cover(b'woooo', 142), MP4Cover(b'woooo', 142))
+        self.assertReallyNotEqual(
+            MP4Cover(b'woooo', 143), MP4Cover(b'woooo', 142))
+        self.assertReallyNotEqual(
+            MP4Cover(b'woooo', 142), MP4Cover(b'wooox', 142))
+
+    def test_cmp_bytes(self):
+        self.assertReallyEqual(MP4Cover(b'woooo'), b"woooo")
+        self.assertReallyNotEqual(MP4Cover(b'woooo'), b"foo")
+        if PY2:
+            self.assertReallyEqual(MP4Cover(b'woooo'), u"woooo")
+            self.assertReallyNotEqual(MP4Cover(b'woooo'), u"foo")
 
 
 class TMP4AudioSampleEntry(TestCase):
@@ -907,7 +1054,10 @@ class TMP4AudioSampleEntry(TestCase):
         fileobj = cBytesIO(atom_data)
         atom = Atom(fileobj)
         entry = AudioSampleEntry(atom, fileobj)
+        self.assertEqual(entry.bitrate, 0)
         self.assertEqual(entry.channels, 1)
+        self.assertEqual(entry.codec, "alac")
+        self.assertEqual(entry.codec_description, "ALAC")
         self.assertEqual(entry.sample_rate, 8000)
 
     def test_alac_2(self):
@@ -922,16 +1072,134 @@ class TMP4AudioSampleEntry(TestCase):
         fileobj = cBytesIO(atom_data)
         atom = Atom(fileobj)
         entry = AudioSampleEntry(atom, fileobj)
-        self.assertEqual(entry.channels, 2)
-        self.assertEqual(entry.sample_rate, 88200)
         self.assertEqual(entry.bitrate, 2437845)
+        self.assertEqual(entry.channels, 2)
+        self.assertEqual(entry.codec, "alac")
+        self.assertEqual(entry.codec_description, "ALAC")
+        self.assertEqual(entry.sample_rate, 88200)
 
-add(TMP4AudioSampleEntry)
+    def test_pce(self):
+        atom_data = (
+            b'\x00\x00\x00dmp4a\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00'
+            b'\x00\x00\x00\x00\x00\x00\x02\x00\x10\x00\x00\x00\x00\xbb\x80'
+            b'\x00\x00\x00\x00\x00@esds\x00\x00\x00\x00\x03\x80\x80\x80/\x00'
+            b'\x00\x00\x04\x80\x80\x80!@\x15\x00\x15\x00\x00\x03\xed\xaa\x00'
+            b'\x03k\x00\x05\x80\x80\x80\x0f+\x01\x88\x02\xc4\x04\x90,\x10\x8c'
+            b'\x80\x00\x00\xed@\x06\x80\x80\x80\x01\x02')
+
+        fileobj = cBytesIO(atom_data)
+        atom = Atom(fileobj)
+        entry = AudioSampleEntry(atom, fileobj)
+
+        self.assertEqual(entry.bitrate, 224000)
+        self.assertEqual(entry.channels, 8)
+        self.assertEqual(entry.codec_description, "AAC LC+SBR")
+        self.assertEqual(entry.codec, "mp4a.40.2")
+        self.assertEqual(entry.sample_rate, 48000)
+        self.assertEqual(entry.sample_size, 16)
+
+    def test_sbr_ps_sig_1(self):
+        atom_data = (
+            b"\x00\x00\x00\\mp4a\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00"
+            b"\x00\x00\x00\x00\x00\x00\x02\x00\x10\x00\x00\x00\x00\xbb\x80\x00"
+            b"\x00\x00\x00\x008esds\x00\x00\x00\x00\x03\x80\x80\x80'\x00\x00"
+            b"\x00\x04\x80\x80\x80\x19@\x15\x00\x03\x00\x00\x00\xe9j\x00\x00"
+            b"\xda\xc0\x05\x80\x80\x80\x07\x13\x08V\xe5\x9dH\x80\x06\x80\x80"
+            b"\x80\x01\x02")
+
+        fileobj = cBytesIO(atom_data)
+        atom = Atom(fileobj)
+        entry = AudioSampleEntry(atom, fileobj)
+
+        self.assertEqual(entry.bitrate, 56000)
+        self.assertEqual(entry.channels, 2)
+        self.assertEqual(entry.codec_description, "AAC LC+SBR+PS")
+        self.assertEqual(entry.codec, "mp4a.40.2")
+        self.assertEqual(entry.sample_rate, 48000)
+        self.assertEqual(entry.sample_size, 16)
+
+        self.assertTrue(isinstance(entry.codec, text_type))
+        self.assertTrue(isinstance(entry.codec_description, text_type))
+
+    def test_als(self):
+        atom_data = (
+            b'\x00\x00\x00\x9dmp4a\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00'
+            b'\x00\x00\x00\x00\x00\x00\x02\x00\x00\x10\x00\x00\x00\x00\x07'
+            b'\xd0\x00\x00\x00\x00\x00yesds\x00\x00\x00\x00\x03k\x00\x00\x00'
+            b'\x04c@\x15\x10\xe7\xe6\x00W\xcbJ\x00W\xcbJ\x05T\xf8\x9e\x00\x0f'
+            b'\xa0\x00ALS\x00\x00\x00\x07\xd0\x00\x00\x0c\t\x01\xff$O\xff\x00'
+            b'g\xff\xfc\x80\x00\x00\x00,\x00\x00\x00\x00RIFF$$0\x00WAVEfmt '
+            b'\x10\x00\x00\x00\x01\x00\x00\x02\xd0\x07\x00\x00\x00@\x1f\x00'
+            b'\x00\x04\x10\x00data\x00$0\x00\xf6\xceF+\x06\x01\x02')
+
+        fileobj = cBytesIO(atom_data)
+        atom = Atom(fileobj)
+        entry = AudioSampleEntry(atom, fileobj)
+
+        self.assertEqual(entry.bitrate, 5753674)
+        self.assertEqual(entry.channels, 512)
+        self.assertEqual(entry.codec_description, "ALS")
+        self.assertEqual(entry.codec, "mp4a.40.36")
+        self.assertEqual(entry.sample_rate, 2000)
+        self.assertEqual(entry.sample_size, 16)
+
+    def test_ac3(self):
+        atom_data = (
+            b'\x00\x00\x00/ac-3\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00'
+            b'\x00\x00\x00\x00\x00\x00\x02\x00\x10\x00\x00\x00\x00V"\x00\x00'
+            b'\x00\x00\x00\x0bdac3R\t\x00')
+
+        fileobj = cBytesIO(atom_data)
+        atom = Atom(fileobj)
+        entry = AudioSampleEntry(atom, fileobj)
+
+        self.assertEqual(entry.bitrate, 128000)
+        self.assertEqual(entry.channels, 1)
+        self.assertEqual(entry.codec_description, "AC-3")
+        self.assertEqual(entry.codec, "ac-3")
+        self.assertEqual(entry.sample_rate, 22050)
+        self.assertEqual(entry.sample_size, 16)
+
+        self.assertTrue(isinstance(entry.codec, text_type))
+        self.assertTrue(isinstance(entry.codec_description, text_type))
+
+    def test_samr(self):
+        # parsing not implemented, values are wrong but at least it loads.
+        # should be Mono 7.95kbps 8KHz
+        atom_data = (
+            b'\x00\x00\x005samr\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00'
+            b'\x00\x00\x00\x00\x00\x00\x02\x00\x10\x00\x00\x00\x00\x1f@\x00'
+            b'\x00\x00\x00\x00\x11damrFFMP\x00\x81\xff\x00\x01')
+
+        fileobj = cBytesIO(atom_data)
+        atom = Atom(fileobj)
+        entry = AudioSampleEntry(atom, fileobj)
+
+        self.assertEqual(entry.bitrate, 0)
+        self.assertEqual(entry.channels, 2)
+        self.assertEqual(entry.codec_description, "SAMR")
+        self.assertEqual(entry.codec, "samr")
+        self.assertEqual(entry.sample_rate, 8000)
+        self.assertEqual(entry.sample_size, 16)
+
+        self.assertTrue(isinstance(entry.codec, text_type))
+        self.assertTrue(isinstance(entry.codec_description, text_type))
+
+    def test_error(self):
+        fileobj = cBytesIO(b"\x00" * 20)
+        atom = Atom(fileobj)
+        self.assertRaises(ASEntryError, AudioSampleEntry, atom, fileobj)
 
 
-NOTFOUND = os.system("tools/notarealprogram 2> %s" % devnull)
+def call_faad(*args):
+    with open(os.devnull, 'wb') as null:
+        return subprocess.call(
+            ["faad"] + list(args),
+            stdout=null, stderr=subprocess.STDOUT)
 
 have_faad = True
-if os.system("faad 2> %s > %s" % (devnull, devnull)) == NOTFOUND:
+try:
+    call_faad()
+except OSError:
     have_faad = False
     print("WARNING: Skipping FAAD reference tests.")
